

interface IConverImageOption {
	sourceImageFormat: "rgba" | "rgb" | "bgr"
	targetShapeFormat?: "nchw" | "nhwc"
	targetColorFormat?: "rgb" | "bgr" | "gray"
	targetNormalize?: { mean: number[], std: number[] }
}

export function convertImage(image: Uint8Array, option?: IConverImageOption) {
	const sourceImageFormat = option?.sourceImageFormat ?? "rgb";
	const targetShapeFormat = option?.targetShapeFormat ?? "nchw";
	const targetColorFormat = option?.targetColorFormat ?? "bgr";
	const targetNormalize = option?.targetNormalize ?? { mean: [127.5], std: [127.5] };

	let rgbReader: (pixel: number) => [number, number, number];
	let pixelCount: number;
	switch (sourceImageFormat) {
		case "bgr":
			rgbReader = pixel => [image[pixel * 3 + 2], image[pixel * 3 + 1], image[pixel * 3 + 0]];
			pixelCount = image.length / 3;
			break;
		case "rgb":
			rgbReader = pixel => [image[pixel * 3 + 0], image[pixel * 3 + 1], image[pixel * 3 + 2]];
			pixelCount = image.length / 3;
			break;
		case "rgba":
			rgbReader = pixel => [image[pixel * 4 + 0], image[pixel * 4 + 1], image[pixel * 4 + 2]];
			pixelCount = image.length / 4;
			break;
	}

	let targetChannelGetter: (stride: number, pixel: number, offset: number) => number;
	switch (targetShapeFormat) {
		case "nchw":
			targetChannelGetter = (stride, pixel, offset) => stride * offset + pixel;
			break;
		case "nhwc":
			targetChannelGetter = (stride, pixel, offset) => stride * pixel + offset;
			break
	}

	let normIndex = 0;
	const normValue = (val: number) => {
		const mean = targetNormalize.mean[normIndex % targetNormalize.mean.length];
		const std = targetNormalize.std[normIndex % targetNormalize.std.length];
		const result = (val - mean) / std;
		++normIndex;
		return result;
	}

	let outBuffer: Float32Array;
	let pixelWriter: (pixel: number, r: number, g: number, b: number) => any
	switch (targetColorFormat) {
		case "rgb":
			outBuffer = new Float32Array(pixelCount * 3);
			pixelWriter = (pixel, r, g, b) => {
				outBuffer[targetChannelGetter(3, pixel, 0)] = normValue(r);
				outBuffer[targetChannelGetter(3, pixel, 1)] = normValue(g);
				outBuffer[targetChannelGetter(3, pixel, 2)] = normValue(b);
			}
			break;
		case "bgr":
			outBuffer = new Float32Array(pixelCount * 3);
			pixelWriter = (pixel, r, g, b) => {
				outBuffer[targetChannelGetter(3, pixel, 0)] = normValue(b);
				outBuffer[targetChannelGetter(3, pixel, 1)] = normValue(g);
				outBuffer[targetChannelGetter(3, pixel, 2)] = normValue(r);
			}
			break;
		case "gray":
			outBuffer = new Float32Array(pixelCount);
			pixelWriter = (pixel, r, g, b) => {
				outBuffer[targetChannelGetter(1, pixel, 0)] = normValue(0.2126 * r + 0.7152 * g + 0.0722 * b);
			}
			break;
	}

	for (let i = 0; i < pixelCount; ++i) {
		const [r, g, b] = rgbReader(i);
		pixelWriter(i, r, g, b);
	}
	return outBuffer;
}